/*
 * Indie Game Jame 2 - example game "framework"
 * 2004/3/15 checker@d6.com
 */


// Add Duck and punch
// tweak jump/leap stuff
// better new object and camera
// etc

// poses - crouch, left and right dodge

// guess not

// ok, how about
//  - game timings and display and such
//  - level progressions - first just stand, then higher target, then objects
//  - maybe a bigger ball (easier or harder), different board, etc...
//  - try to add arms, crouch, etc...

// presentation
//  - add some sounds, eh?

// real high scores
//  - track lots of high-scores (and high-score file, i guess)
//    - longest air time
//    - most jumps w/o a dab
//    - longest balance
//    - per-level high-scores

// relax model more
//  - add procedural arm movement
//  - try and make auto-crouch work/auto-dodge work

// misc
//  - cleanup gamestate logic mess

// RIGHT NOW
//  score fix after Select/Craziness
//  fix throwing

// go back and work on other two games

#include "shared/core/first.h"

// windows crap
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <windowsx.h>

// system
#include <assert.h>
#define _USE_MATH_DEFINES
#include <math.h>

// gl stuff
#include <gl/gl.h>
#include "glutils.h"

// physics stuff
#include "shared/physics/2d/2d_body_hull.h"
#include "shared/physics/2d/2d_body_list.h"
#include "shared/physics/2d/2d_body_box.h"
#include "shared/physics/2d/2d_body_sphere.h"
#include "shared/physics/2d/rigid/2d_rigid_sim.h"
#include "shared/physics/2d/rigid/2d_rigid_drag_force.h"
#include "shared/physics/2d/rigid/2d_rigid_gravity_force.h"
#include "shared/physics/2d/rigid/2d_rigid_spring_force.h"
#include "shared/physics/2d/rigid/2d_rigid_control_force.h"
#include "shared/input/joystick.h"
using namespace ajb;
using namespace Physics::D2;
using namespace Physics::D2::Rigid;

// misc
#include "utils.h"
#include <vector>
using namespace std;

// media lib
#include "music.h"
#include "zsound.h"
#include "bitmap.h"

#include "property_system.h"
#include "property_system_editor.h"
#include "editor.h"

/********************************* globals **********************************/

// window/camera stuff
char const AppName[] = "IGJ2 - balance board";
int WindowWidth = 800;
int WindowHeight = 600;
bool AppQuit = false;  // stuff this to kill yourself
float FOVY = 45.0f, ZNear = 0.1f, ZFar = 1000.0f;  // whatever, we're 2D
float AspectRatio = float(WindowWidth)/float(WindowHeight); // updated in Resize
float CameraAngleRadians = 0.0, CameraX = 0.0;
Vec2 WorldFromWindowCoords( Vec2 window );
Vec2 WindowFromWorldCoords( Vec2 world );

int game_paused=0;
int draw_debugging_info=0;

// win32 stuff
HWND WindowHandle;
HDC WindowDC;

// gl stuff
HGLRC WindowRC;

// functions for you to play in
void Resize( int w, int h );
void Initialize( void );
void Uninitialize( void );
void Idle( void );
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

// system
Music music;

// physics stuff

// needed for game consistency on different machines
#include "../../utils/misc/framerate.h"

Sim simulator;
Drag_Force    global_drag(.005f, .005f);    //a little drag helps with stabiplity
Gravity_Force gravity(Vec2(0.0f, -20.0f));
Rigid_Body    world_body(make_Body_Null(),Vec2(0,0),0,0,0);

char *CurLevelFileName="guy.physdb";
float CameraZ=100.0f, CameraY=20.f;

PROP_PARSE_THIS_FILE;
PROP_STRUCT struct global_properties
{
    // This is the gravity
   float Gravity;
   
   float MouseSpringKs;
   float MouseSpringKd;
   float MouseSpringRestLength;
   
   float ForceMul;
   float ForceJump;
   float ForceFoot;

   int Foo;
};
static global_properties Globals;
static property *GlobalProperties;
static property_system_editor *GlobalsEditor;

static bool MovingCamera = false;

enum game_object_type {
   FootType,
   ArmType,
   BoardType,
   WorldType,
   BallType,
   HeadType,
   RiderType,
   DropType,
   UnknownType,
};

static float typecol[][3]=
{
   {0,1,0},
   {0,1,0},
   {.6f,0,.6f},
   {0,0,1},
   {.75f,.4f,0},
   {0,.7f,0},
   {0,1,0},
   {1,0,0},
   {0,1,0},
};

// game stuff
struct game_object {
   Rigid_Body *rb;
   
   game_object( char *texture_file , Rigid_Body *rbInit ) :
      rb(rbInit)
   {
      assert(rb);
      simulator.add(*rb);
      rb->m_pAppData=(void *)this;
   }
   
   ~game_object( void )
   {
      delete rb;
   }
   
   game_object_type Type;
};

static vector<game_object*> objs;
static vector<Binary_Force *> forces;

static Vec2 window_mouse, world_mouse;
static Joystick PS2Controller;

static Binary_Force *spring = 0;

// attempt to track the key objects in the world
static game_object  *pFoot[2]={0,0};
static game_object  *pArm[2]={0,0};
static Binary_Force *pPtFoot[2]={0,0};
static int           ptleg_active[2]={1,1};
static game_object  *pWorld=0, *pBoard=0, *pBall=0;
static float         world_top=0;

void initFoot(game_object *Obj, int lr) { pFoot[lr]=Obj; }
void initArm (game_object *Obj, int lr) { pArm[lr]=Obj; }

typedef void obj_type_init(game_object *Obj, int lr);

char             *typenames[]={"Foot","Arm","Board","World","Ball","Rider","Drop"};
game_object_type  typetypes[]={FootType,ArmType,BoardType,WorldType,BallType,RiderType,DropType};
int               typeuselr[]={1,1,0,0,0,0,0};
obj_type_init    *typeinit[]={initFoot,initArm,0,0,0,0,0};
game_object     **typeptrs[]={0,0,&pBoard,&pWorld,&pBall,0,0};
#define NUM_TYPES_LOADED (sizeof(typenames)/sizeof(typenames[0]))

// why couldnt i get this working?  i suck
//game_object    *(*typeptrslr[])={pFoot,pHand,0,0,0,0};

#define MAX_POSES  8
#define MAX_FORCES 60
Binary_Force *pose_forces[MAX_POSES][MAX_FORCES];
int pose_force_cnt=0;

static char *pose_files[]={"guy.physdb","crouchguy.physdb"};
#define NUM_POSES (sizeof(pose_files)/sizeof(pose_files[0]))

void LoadObjectsFromFile(char *fname)
{
    int num_rb=0, num_force=0;
    Rigid_Body **rb_list;
    Binary_Force **force_list;

    loadPhysicsFile(fname, 0, 0, 1.0/16, 1.0/16, // 0, 0, 1.0, 1.0,
                    &rb_list, &num_rb, &force_list, &num_force, &world_body);

    for (int i=0; i < num_rb; ++i)
    {
       game_object *Obj = new game_object("tmp",rb_list[i]);
       objs.push_back(Obj);
       Obj->Type=UnknownType;

       for (int j=0; j<NUM_TYPES_LOADED; j++)
       {
          char *cmpstr=typenames[j];
          int   cmplen=(int)strlen(cmpstr);
          if (strnicmp(Obj->rb->name().c_str(),cmpstr,cmplen)==0)
          {
             int lr=-1;
             if (typeuselr[j])
             {
                if (strnicmp(Obj->rb->name().c_str()+cmplen,"Left",strlen("Left"))==0)
                   lr=0;
                else if (strnicmp(Obj->rb->name().c_str()+cmplen,"Right",strlen("Right"))==0)
                   lr=1;
                else
                   assert(0);
             }
             if (typeptrs[j]!=0)
                *typeptrs[j]=Obj;
             Obj->Type=typetypes[j];
             if (typeinit[j])
                (*typeinit[j])(Obj,lr);
             break;
          }
       }
    }

    pose_force_cnt=num_force;
    for (int i=0; i < num_force; ++i)
    {
       if (strnicmp(force_list[i]->name().c_str(),"PtFoot",strlen("PtFoot"))==0)
       {
          if (strnicmp(force_list[i]->name().c_str(),"PtFootLeft",strlen("PtFootLeft"))==0)
             pPtFoot[0]=force_list[i];
          else if (strnicmp(force_list[i]->name().c_str(),"PtFootRight",strlen("PtFootRight"))==0)
             pPtFoot[1]=force_list[i];
          else
             continue;  // something wrong here
       }  
       forces.push_back(force_list[i]);
       simulator.add(*force_list[i]);

       pose_forces[0][i]=force_list[i];
    }

}

// fills a vec2 with x min and y max
void Get_MinMaxY(game_object *pObj, Vec2 &minmax)
{
   // find the low point of the world
   Body     *pBody = dynamic_cast<Body*>(pObj->rb->m_Pbody.get());
   Body_Box *pBox  = dynamic_cast<Body_Box*>(pBody);
   minmax.set_x(+1000000.0f);
   minmax.set_y(-1000000.0f);
   for (int i=0; i<pBox->num_verts();i++)
   {
      if (pBox->vert(i).y()>minmax.y())
         minmax.set_y(pBox->vert(i).y());
      if (pBox->vert(i).y()<minmax.x())
         minmax.set_x(pBox->vert(i).y());         
   }
}

void LoadPoseList(int pose_id)
{
    int num_rb=0, num_force=0;
    Rigid_Body **rb_list;
    Binary_Force **force_list;

    if (pose_id>=NUM_POSES)
       return;

    loadPhysicsFile(pose_files[pose_id], 0, 0, 1.0/16, 1.0/16, // 0, 0, 1.0, 1.0,
                    &rb_list, &num_rb, &force_list, &num_force, &world_body);

    assert(pose_force_cnt==num_force);

    for (int i=0; i<num_force; i++)
       pose_forces[pose_id][i]=force_list[i];
}

GLuint LoadTexture(char *Filename)
{
    GLuint texture;
    
    Bitmap_Load_Result result;
    verify(LoadBitmapByExtension(Filename,&result));

    // gl stuff
    texture = gl_texture_from_bitmap(result.data, result.width, result.height,
                                     result.bytes_per_pixel);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    // delete the bitmap
    delete [] result.data;

    return(texture);
}

void
glBody1v(const game_object &o, Body *body, bool Solid)
{
    if(body)
    {
        {
            Body_Hull *p = dynamic_cast<Body_Hull*>(body);
            if(p)
            {
                glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
                for(int i = 0;i<p->num_verts();++i)
                {
                    Vec2 worldspace = p->vert(i);
                    glVertex2f(worldspace.x(),worldspace.y());
                }
                glEnd();
            }
        }

        {
            Body_Box *p = dynamic_cast<Body_Box*>(body);
            if(p)
            {
                glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
                for(int i = 0;i<p->num_verts();++i)
                {
                    Vec2 worldspace = p->vert(i);
                    if (Solid)
                       glColor3f(typecol[o.Type][0],typecol[o.Type][1],typecol[o.Type][2]);
                    glVertex2f(worldspace.x(),worldspace.y());
                }
                glEnd();
            }
        }

        {
            Body_Sphere *p = dynamic_cast<Body_Sphere*>(body);
            if(p)
            {
                static float vcol[4][3]={{.4f,.2f,0},{.63f,.36f,0},{.82f,.43f,0},{.57f,.39f,0}};
                glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
                int total_steps = 16;
                Vec2 center = p->x();
                if(Solid)
                {
                   glColor3f(vcol[1][0],vcol[1][1],vcol[1][2]);
                   glVertex2f(center.x(), center.y());
                }
                for(int a = 0;a<total_steps;++a)
                {
                    float Angle = 2.0f * 3.14159265358979323846f * ((float)a / (float)(total_steps - 1));
                    float Radius = p->r();
                    Angle+=o.rb->O.z();
                    Vec2 arm(Radius*cos(Angle), Radius*sin(Angle));
                    arm += center;
                    if (Solid)
                       glColor3f(vcol[a&3][0],vcol[a&3][1],vcol[a&3][2]);
                    glVertex2f(arm.x(), arm.y());
                }
                glEnd();
            }
        }

        Body_List *l = dynamic_cast<Body_List*>(body);
        if(l)
        {
            {for(int Index = 0;
                 Index < l->num_bodies();
                 ++Index)
            {
               glBody1v(o,&l->body(Index), Solid);
            }}
        }
    }
}

void
glForce1v(Binary_Force *Force)
{
    if(Force)
    {
        glBegin(GL_LINES);
        glVertex2f(Force->m_a.x_worldS().x(), Force->m_a.x_worldS().y());
        glVertex2f(Force->m_b.x_worldS().x(), Force->m_b.x_worldS().y());
        glEnd();
    }
}

void
ClearMouseSpring(void)
{
    if(spring)
    {
        simulator.remove(*spring);
        delete spring;
        spring = 0;
    }
}

void
ClearAll(void)
{
    ClearMouseSpring();
    
    for(int i = 0;i < objs.size();++i)
    {
        assert(objs[i]);
        simulator.remove(*objs[i]->rb);
    }

    for(int i = 0;i < forces.size();++i)
    {
        assert(forces[i]);
        simulator.remove(*forces[i]);
    } 

    forces.clear();
    objs.clear();
}

void LoadSingleObject(char *fname, float x=0, float y=0)
{
   int num_rb=0, num_force=0;
   Rigid_Body **rb_list;
   Binary_Force **force_list;

   loadPhysicsFile(fname,x,y,1.0/16,1.0/16,
                    &rb_list, &num_rb, &force_list, &num_force, &world_body);

}

////////////////
// utils
float frand(float lo=0, float hi=1.0f)
{
   float zt1=((float)(rand()&0x7fff))/(float)0x7fff;
   return lo + ((hi-lo)*zt1);
}

// x will be window/2
void DrawXCenteredString(float y, char *string, int font_id=-1)
{
   float x=(WindowWidth/2)-(GetTextWidth(string,font_id)/2);
   DrawString(x,y,string,font_id);
}

/****************************************************************************/
/* Game Stuff                                                               */
// player controlled forces on the legs
static Constant_Force *leg_forces[2]={0,0};

// Make timing sane, deal w/Pause

// Misc Game State
int running_live=0;
int drop_running=0;

int score=0;
int total_score=0;

int cur_dab=0;
int cur_air=0;
int cur_jumps=0;       // how many airs between dabs this time
int last_air=0;

int active_dropobjs=0;

int head_plant=0;
int body_plant=0;
int cur_fell=0;
int cur_pose=0;

int game_lives=3;

// gonna have to debounce this right!
int dab_count=0;

// Scoring system
//  - depending on seconds since last dab, you score 1-10 points every second
//  - if in the air, each second is worth x5
//  - subsequent jumps w/o a dab are worth x10, x15, etc
//  - while in the air, point counter goes up, but if you dab or miss, pts arent scored
int cur_score_run=0;

double game_time=0;    // what is the current game time
double drop_timer=0;   // at what game time will we drop/throw next
double last_dab=0;     // at what game time did player last dab
double last_score=0;   // at what game time did player last score
double run_time=0;     // time of current run - sink to best_time 
double best_time=0;    // best continuous time in this session
double takeoff_time=0; // when last jump started
double raw_time=0;     // for hookup of deltat in frames

double last_dab_sound=0;

// high scores
int   hs_points=0;
int   hs_jumps=0;
float hs_best_time=0.f;

// sinking points in a pretty way, ideally
int   sink_score=0;
float sink_timer=0;

// minmax locations for detecting when board has left world...
Vec2 World_MinMaxY;
Vec2 Board_MinMaxY;

// setup game mode
void Game_Init(void)
{
   game_time=drop_timer=last_dab=best_time=last_dab_sound=run_time=takeoff_time=last_score=0.f;
   score=cur_dab=drop_running=cur_fell=head_plant=cur_score_run=cur_jumps=last_air=dab_count=0;
   raw_time=GetTime();

   running_live=1;

   ClearAll();
   LoadObjectsFromFile(CurLevelFileName);

//   for (int i=0; i<NUM_POSES; i++)
//      LoadPoseList(i);

   Get_MinMaxY(pWorld,World_MinMaxY);
}

// helper function
float  game_throw_mul=2.f;

game_object *throw_obj(const Vec2& src_pt, const Vec2& targ_pt, const int shape, const float size)
{  // make_Body_Sphere(size);
   game_object *pObj=NULL;

   Vec2 Vel=(targ_pt-src_pt)*game_throw_mul;
   Vel.set_y(Vel.y()+16.f);

   switch (shape)
   {
      case 0:
         pObj = new game_object("tmp",new Rigid_Body(make_Body_Sphere(size), src_pt, 0, 1, 1, Vel));
         break;
      case 1:
         pObj = new game_object("tmp",new Rigid_Body(make_Body_Box(Vec2(frand(-size,-size/2),frand(-size,-size/2)),
                                                                   Vec2(frand(size,size/2),frand(size,size/2))), 
                                                     src_pt, 0, 1, 1, Vel, frand(-2,2)));
         break;
   }
   if (pObj)
   {
      pObj->Type=DropType;
      pObj->rb->k_friction=0.005f;
      
      // set velocity

      objs.push_back(pObj);
      active_dropobjs++;
   }
   return pObj;
}

enum game_states 
{
   kGameStateTitle,
   kGameStateInLevel,
   kGameStateLevelSwitch,
   kGameStateLevelFailed,
   kGameStateGameOver,
   kGameStateScoreSummary,
   kGameStateFreeForm,
   kGameStateNumStates=kGameStateFreeForm+1
};
game_states cur_game_state=kGameStateTitle;
double game_state_timer=0;

struct sLevelInfo 
{
   char  fname[32];
   char  text[120];
   float time_limit;
   int   score_goal;
   int   dab_max;
   float dodge_freq;
   float ground_spd;
   float ground_flip;
   float throw_power;
};

int    cur_level=0;
sLevelInfo gLevelData[] =
{
   { "level0.physdb", "Stay in the world for 60 seconds", 60.f, 0, 0, 0, 0, 0, 0},
   { "level1.physdb", "60 seconds, under 10 dabs", 60.f, 0, 10, 0, .05f, 6.f, 0 },
   { "level2.physdb", "60 seconds, under 10 dabs, over 200 points, watch out", 60.f, 200, 10, 20.f, .025f, 6.f, 1.5f },
   { "level3.physdb", "90 seconds, under 10 dabs, watch the ground", 90.f, 0, 10, 0.f, .25f, 1.5f, 3.f },
   { "level4.physdb", "20 dabs, as long as you want, go for points", 0.f, 0, 20, 10.f, .1f, 5.f, 3.f },
};
#define NUM_LEVELS (sizeof(gLevelData)/sizeof(gLevelData[0]))

// next will be rotating the ground plane
char *failure_strings[] =
{
   "Too Many Dabs, Try Again",
   "Nice Try, but you need more points",
   "You fell off the world"
};
enum failure_codes 
{
   kFailDabs,
   kFailPoints,
   kFailFell,
   kFailDidntFail, // take that, oh pathetic science of logic
};
failure_codes gFailureCode=kFailDidntFail;

int game_won=0;

void Game_State_Enter(game_states new_state)
{  // do any setup/magic
   cur_game_state=new_state;
   game_state_timer=0.f;
}

void Goto_State_Fail(failure_codes whyFailed)
{
   Game_State_Enter(kGameStateLevelFailed);
   gFailureCode=whyFailed;
}

void Switch_Level(int new_level)
{
   if ((new_level>=NUM_LEVELS)||(new_level<0))
      return;
   cur_level=new_level;
   CurLevelFileName=&gLevelData[cur_level].fname[0];
   Game_Init();
}

// when to next throw something in...
void set_drop_timer(void)
{
   float dodge_freq=gLevelData[cur_level].dodge_freq;
   float next_drop_delta=dodge_freq-((float)game_time/10.0f);
   if (next_drop_delta<2.f) next_drop_delta=2.f;
   drop_timer=game_time+next_drop_delta;
}

#define INITIAL_LIVES 3
#define TIME_PER_MODE 5.f

int Timer_Finished(double deltat, int allow_click_through=0)
{
   if (game_state_timer==0)
      game_state_timer=TIME_PER_MODE;
   if (game_state_timer>0.f)
   {
      game_state_timer-=deltat;
      if (game_state_timer<=0.f)
         return 1;
      if (allow_click_through)
         if (PS2Controller.button_went_down(Joystick::BUTTON_START)||
             PS2Controller.button_went_down(Joystick::BUTTON_RIGHT_D))
            return 1;
   }
   return 0;
}

void Game_State_Manager(double deltat)
{
   switch (cur_game_state)
   {
      case kGameStateFreeForm:
         break;
      case kGameStateTitle:
         total_score=game_won=0;
         game_lives=INITIAL_LIVES;
         if (PS2Controller.button_went_down(Joystick::BUTTON_START)||
             PS2Controller.button_went_down(Joystick::BUTTON_RIGHT_D))
         {
            Game_State_Enter(kGameStateInLevel);
            Switch_Level(0);
         }
         else if (PS2Controller.button_went_down(Joystick::BUTTON_SELECT))
            Game_State_Enter(kGameStateFreeForm);
         break;
      case kGameStateInLevel:
         // do time based level changing
         sLevelInfo *pLevel;
         pLevel=&gLevelData[cur_level];
         if (pLevel->time_limit>0 && game_time>pLevel->time_limit)
         {  // level time out, you are done with it...
            if (pLevel->score_goal==0 || score>=pLevel->score_goal)
               Game_State_Enter(kGameStateLevelSwitch);
            else
               Goto_State_Fail(kFailPoints);
         }
         if (pLevel->dab_max>0 && dab_count>=pLevel->dab_max)
            Goto_State_Fail(kFailDabs);
         if (cur_fell)
            Goto_State_Fail(kFailFell);
         if (pLevel->ground_spd)  // play with rotation, eh?
         {
            int dir=(int)((game_time+(pLevel->ground_flip/2))/pLevel->ground_flip);
            pWorld->rb->w.set_z(dir&1?pLevel->ground_spd:-pLevel->ground_spd);
         }
         if (pLevel->dodge_freq)
         {
            if (drop_timer==0)
               drop_timer=frand(gLevelData[cur_level].dodge_freq/3,gLevelData[cur_level].dodge_freq);
            if ((drop_timer!=0)&&(game_time>drop_timer))
            {
               throw_obj(Vec2(frand()>0.5f?-40.f:40.f,frand(0.0f,60.f)),
                         Vec2(frand()>0.5f?-10.f:10.f,frand(10.0f,40.f)),
                         frand()>0.9?1:0,1.f);
               set_drop_timer();
            }
         }
         break;
      case kGameStateLevelSwitch:
         if (Timer_Finished(deltat,1))
         {
            total_score+=score;
            if (cur_level+1>=NUM_LEVELS)  // you have won!
            {
               Game_State_Enter(kGameStateGameOver);
               game_won=1;
            }
            else
            {
               Switch_Level(cur_level+1);
               Game_State_Enter(kGameStateInLevel);
            }
         }
         break;
      case kGameStateLevelFailed:
         if (Timer_Finished(deltat,1))
         {
            if (--game_lives>0)
            {
               Switch_Level(cur_level);
               Game_State_Enter(kGameStateInLevel);
            }
            else
               Game_State_Enter(kGameStateGameOver);
         }
         break;
      case kGameStateGameOver:
         if (Timer_Finished(deltat,0))
            Game_State_Enter(kGameStateTitle);
         break;
   }
}

void flip_foot_state(int foot)
{
   if (ptleg_active[foot])
   {
      simulator.remove(*pPtFoot[foot]);
      ptleg_active[foot]=0;
      leg_forces[foot]->m_f+=Vec2(0,Globals.ForceFoot);
   }
   else
   {
      simulator.add(*pPtFoot[foot]);
      ptleg_active[foot]=1;
      leg_forces[foot]->m_f+=Vec2(0,-Globals.ForceFoot/4);
   }
}

#if 0
// move each elements of the forces list towards the pose
void smooth_towards_pose(void)
{
   for (i=0; i<pose_force_cnt; i++)
   {
   }
}
#endif

void switch_pose(int pose_id)
{
   int i;

   if (pose_id>=NUM_POSES)
      return;
   if (pose_forces[pose_id][0]==0)
      return; // oh well

   for(i=0; i < forces.size() ;++i)
   {
      assert(forces[i]);
      simulator.remove(*forces[i]);
   }
   forces.clear();
   for (i=0; i<pose_force_cnt; i++)
   {
      forces.push_back(pose_forces[pose_id][i]);
      simulator.add(*pose_forces[pose_id][i]);
   }
}

int CheckPairForTypes(game_object *pA, game_object *pB, game_object_type TA, game_object_type TB)
 { return (pA->Type==TA&&pB->Type==TB) || (pA->Type==TB&&pB->Type==TA); }

void Game_Tick(void)
{
   if (cur_game_state==kGameStateFreeForm)
   {
      if (PS2Controller.button_went_down(Joystick::BUTTON_SELECT))
         Game_Init();
      if (PS2Controller.button_went_down(Joystick::BUTTON_START))
         game_paused=!game_paused;
      
      // debug/control/test code
      if (PS2Controller.button_went_down(Joystick::BUTTON_RIGHT_R))
         draw_debugging_info=!draw_debugging_info;
      if (PS2Controller.button_went_down(Joystick::BUTTON_RIGHT_L))
         Game_State_Enter((game_states)((cur_game_state+1)%(int)kGameStateNumStates));
      
      // neither of these really works right now
      if (PS2Controller.button_went_down(Joystick::BUTTON_RIGHT_U))
      {
         drop_running=!drop_running;
         if (drop_running) 
            set_drop_timer();
      }
      if (PS2Controller.button_went_down(Joystick::BUTTON_RIGHT_D))
      {
         cur_pose=(cur_pose+1)%NUM_POSES;
         switch_pose(cur_pose);
      }

   }
   else if (cur_game_state!=kGameStateTitle)
   {
      if (PS2Controller.button_went_down(Joystick::BUTTON_START))
         game_paused=!game_paused;
      if (PS2Controller.button_went_down(Joystick::BUTTON_SELECT))
         Game_Init();
   }

   double now_time=GetTime();
   double deltat=now_time-raw_time;
   raw_time=now_time;

   Game_State_Manager(deltat);

   // now gameplay (tm)
   if ((!running_live)||(game_paused))
      return;

   game_time+=deltat;

   Get_MinMaxY(pBoard,Board_MinMaxY);

   // list of guys to delete after we finish collision analysis
   game_object *delete_list[16];
   int delete_count=0;
   
   // analysis of the players current state
   cur_dab=body_plant=head_plant=body_plant=0; cur_air=1;
   for (Contact_List::const_iterator cit=simulator.m_contacts.begin(); 
        cit != simulator.m_contacts.end(); ++cit) 
   {
      game_object *pObjA = (game_object *) (((Rigid_Body *)(cit->m_Prb_a))->m_pAppData);
      game_object *pObjB = (game_object *) (((Rigid_Body *)(cit->m_Prb_b))->m_pAppData);
      cur_dab|=CheckPairForTypes(pObjA,pObjB,BoardType,WorldType);
      if (CheckPairForTypes(pObjA,pObjB,BoardType,BallType))
         cur_air=0;
      if (CheckPairForTypes(pObjA,pObjB,HeadType,WorldType))
         head_plant=1;
      if (CheckPairForTypes(pObjA,pObjB,RiderType,WorldType))
         body_plant=1;
      if (CheckPairForTypes(pObjA,pObjB,DropType,WorldType))
         delete_list[delete_count++]=(pObjA->Type==DropType?pObjA:pObjB);
   }
   // boardy Max (y) is < world Min (x), board is below the world
   if (Board_MinMaxY.y() < World_MinMaxY.x()) 
   {
      cur_fell=1;
//      zsoundPlay("content/sounds/us bodyfall 01.wav");
   }

   // do deletes - though first debounce the stupid list
   for (int j=0; j<delete_count; j++)
   {
      int skip=0;
      for (int k=j+1; k<delete_count; k++)
         if (delete_list[j]==delete_list[k])
            skip=1;
      if (skip)
         continue;
      simulator.remove(*(delete_list[j]->rb));
      // i cant believe you have to do this to remove from a vector, jesus
      objs.erase(std::find(objs.begin(),objs.end(),delete_list[j]));
      active_dropobjs--;
      delete delete_list[j];
   }

   // here we need to update all our times and such
   if (cur_fell)  // oops, bad news
   {
   }
   else if (cur_dab==0 && body_plant==0 && head_plant==0)
   {
      if (game_time-last_dab>run_time)
         run_time=game_time-last_dab;   // new "long run" for us
      if (last_air&&(cur_air==0))
      {  // were in the air, now not dabbed and not in the air = ball
         score+=cur_score_run;
         sink_score=cur_score_run;
         sink_timer=5.f;
         cur_score_run=0;
         if (game_time>takeoff_time+0.25f)
         {  // make jump landing sound
            cur_jumps++;
            zsoundPlay("content/sounds/wood_misc_1.wav");
         }
      }
      if (game_time>last_score+1.f)
      {
         int pts=((int)((game_time-last_dab)/5))+1;
         if (pts>20) pts=20;
         if (cur_air)
         {
            pts*=5*(cur_jumps+1);
            cur_score_run+=pts;
         }
         else
            score+=pts;
         last_score+=1.f;
      }
      if ((cur_air==1)&&(last_air==0))
         takeoff_time=game_time;
   }
   else // we have dabbed
   {
      cur_jumps=cur_air=0;  // if dabbing, not airborne
      if (game_time>last_dab+0.25f)
         if (game_time>last_dab_sound+1.f)
         {
            last_dab_sound=game_time;
            zsoundPlay("content/sounds/wood_4.wav");
            dab_count++;
         }
      last_dab=game_time;
   }
   last_air=cur_air;

   // Apply Forces from the sticks
   if ((pFoot[0]!=0)&&(pFoot[1]!=0)) // we have legs
   {
      if (leg_forces[0]==0)  // create the leg forces
      {                        // need to check size of RB for real, eh?
         leg_forces[0] = new Constant_Force(Handle(*(pFoot[0]->rb),Vec2(0.0f,0.0f)));
         simulator.add(*leg_forces[0]);
         leg_forces[1] = new Constant_Force(Handle(*(pFoot[1]->rb),Vec2(0.0f,0.0f)));
         simulator.add(*leg_forces[1]);
      }

      Vec2 Left=PS2Controller.left_stick();
      Left.set_x((Left.x()<0?-1:1)*Left.x()*Left.x());
      Left.set_y((Left.y()<0?-1:1)*Left.y()*Left.y());
      if (Left.y()>0) 
      {
         Left.set_y(0);
         pFoot[0]->rb->k_gravity=0;
      }
      else pFoot[0]->rb->k_gravity=1.0f;
      Vec2 Right=PS2Controller.right_stick();
      Right.set_x((Right.x()<0?-1:1)*Right.x()*Right.x());
      Right.set_y((Right.y()<0?-1:1)*Right.y()*Right.y());
      if (Right.y()>0)
      {
         Right.set_y(0);
         pFoot[1]->rb->k_gravity=0;
      }
      else pFoot[1]->rb->k_gravity=1.0f;
      leg_forces[0]->m_f=Left*Globals.ForceMul;
      leg_forces[1]->m_f=Right*Globals.ForceMul;
   }

   if (PS2Controller.button_went_down(Joystick::BUTTON_LSTICK))
      leg_forces[0]->m_f+=Vec2(0,Globals.ForceJump);
   if (PS2Controller.button_went_down(Joystick::BUTTON_RSTICK))
      leg_forces[1]->m_f+=Vec2(0,Globals.ForceJump);

   if (PS2Controller.button_went_down(Joystick::BUTTON_SHOULDER_L2))
      flip_foot_state(0);
   if (PS2Controller.button_went_down(Joystick::BUTTON_SHOULDER_R2))
      flip_foot_state(1);

   // hmmm, some gameplay, maybe?
   if (drop_running)
      if ((drop_timer!=0)&&(game_time>drop_timer))
      {
         throw_obj(Vec2(frand()>0.5f?-40.f:40.f,frand(0.0f,60.f)),
                   Vec2(frand()>0.5f?-10.f:10.f,frand(10.0f,40.f)),
                   frand()>0.9?1:0,1.f);
         set_drop_timer();
      }
}

// game specific draw code
void Game_Draw(void)
{
   char buf[120];

   if (draw_debugging_info)
   {
      if (cur_dab==1)
         DrawString(10.f,60.f,"Dab");
      if (cur_air==1)
         DrawString(10.f,70.f,"Air");
      if (cur_fell==1)
         DrawString(10.f,80.f,"Fell");
      if (head_plant==1)
         DrawString(10.f,50.f,"Head");
      sprintf(buf,"Score %d",score);
      DrawString(30.f,80.f,buf);
      double cur_len=(cur_dab?0:game_time-last_dab);
      sprintf(buf,"Len %2.2f Best %2.2f Active %d",(float)cur_len,(float)run_time,active_dropobjs);
      DrawString(30.f,70.f,buf);
      sprintf(buf,"Run %d Jumps %d SinkT %2.2f SinkP %d\n",cur_score_run,cur_jumps,sink_timer,sink_score);
      DrawString(30.f,60.f,buf);
   }

   // Do "Real Game" Display here
   SetDefFont(1);  // into big game font_id

   float cur_len, x, y, delta_y;

   switch (cur_game_state)
   {
      case kGameStateFreeForm:
         DrawXCenteredString(570.f,"FreeForm Mode",0);
         break;
      case kGameStateTitle:
         DrawXCenteredString(480.f,"Balance", 2);
         DrawXCenteredString(400.f,"Press X to Continue",1);
         DrawXCenteredString(200.f,"Left/Right Stick apply force to Left/Right foot",1);
         DrawXCenteredString(150.f,"click Stick to push down hard w/that foot",1);
         break;
      case kGameStateInLevel:
         x=GetTextWidth(buf,0);
         sprintf(buf,"Level %d: %s",cur_level,gLevelData[cur_level].text);
         DrawXCenteredString(550.f,buf,0);

         delta_y=GetTextHeight()+2.f;
         y=WindowHeight-30.f;
         x=WindowWidth-GetTextWidth("TestStringLen");

         // now show pretty score displays
         sprintf(buf,"Score %d",score);
         DrawString(x,y,buf); y-=delta_y;
         cur_len=(float)(cur_dab?0:game_time-last_dab);
         sprintf(buf,"Clean %2.2f",cur_len);
         DrawString(x,y,buf); y-=delta_y;
         sprintf(buf,"Time %2.2f",game_time);
         DrawString(x,y,buf); y-=delta_y;
         sprintf(buf,"Dabs %d",dab_count);
         DrawString(x,y,buf); y-=delta_y;

         // show any current combo if in air
         if (cur_air)
         {
            sprintf(buf," Jump %d",cur_jumps+1);
            DrawString(x,y,buf); y-=delta_y;
            sprintf(buf," Trick %d x%d",cur_score_run,5*(cur_jumps+1));
            DrawString(x,y,buf); y-=delta_y;
         }

         // show the Sink Score here...

         break;

      case kGameStateLevelSwitch:
         DrawXCenteredString(15,"Congratulations",2);
         sprintf(buf,"Level %d - %d Points",cur_level,score);
         DrawXCenteredString(400,buf,1);
         sprintf(buf,"You now have %d total",score+total_score);
         DrawXCenteredString(360,buf,1);
         break;

      case kGameStateLevelFailed:
         DrawXCenteredString(300,failure_strings[(int)gFailureCode],1);
         break;
            
      case kGameStateGameOver:
         if (game_won)
         {
            DrawXCenteredString(200,"You Win",2);
            DrawXCenteredString(180,"(nice)",0);
         }
         else
         {
            DrawXCenteredString(200,"Out of Lives, Sorry",1);
            DrawXCenteredString(150,"Try Again",1);
         }
         break;

      case kGameStateScoreSummary:
         
         break;
   }

   SetDefFont(0);  // back to normal
}

void Game_Key(WPARAM wParam)
{
   switch (wParam)
   {
      case '0':  // debug draw stuff
         draw_debugging_info=!draw_debugging_info;         
         break;
      case '+':  // next level
         Switch_Level((cur_level+1)%NUM_LEVELS);
         break;
      case '-':  // back a level
         Switch_Level((cur_level+NUM_LEVELS-1)%NUM_LEVELS);
         break;
      case 'N':  // next game state
         Game_State_Enter((game_states)((cur_game_state+1)%(int)kGameStateNumStates));
         break;
   }
}

/******************************** Initialize ********************************/
/* the window and graphics subsystem are set up but still hidden            */
void Initialize( void )
{
    PS2Controller.init();

    Globals.Gravity = -30.0f;

    Globals.MouseSpringKs = 128.0f;
    Globals.MouseSpringKd = 32.0f;
    Globals.MouseSpringRestLength = 0.0f;

    Globals.ForceMul  = 400.0f;
    Globals.ForceJump =-25000.0f;
    Globals.ForceFoot = 3000.0f;

    GlobalProperties = CloneProperty(PROP_LOCAL_STRUCT(global_properties));
    ConvertStructToProperty(PROP_LOCAL_STRUCT(global_properties), &Globals,
                            GlobalProperties);
    GlobalsEditor = OpenPropertySystemEditor("Globals", 850, 50, 400, 400);
    EditPropertiesInPlace(GlobalsEditor, GlobalProperties);
    
    GetTime(); // throw away initial 0.0 so we're seconds-from-startup

    // gl stuff
    glClearColor(1.0f,1.0f,1.0f,0.0f);
    glEnable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
    glEnable(GL_DEPTH_TEST);
    glShadeModel(GL_SMOOTH);

    InitializeFrameRate(60.0f);
    
    // init our view
    Resize(WindowWidth,WindowHeight);

    // music init
    music.init();
    music.set_subdirectory("content/music/");
    music.setupnotes();
    music.loadseg("background", "simple_loop.sgt", MUSIC_TRACK_MAIN);

    // sound init
    zsoundInit(WindowHandle);

    // initialize the physics
    simulator.add(global_drag);
    simulator.add(gravity);
    simulator.add(world_body);
}

/*********************************** Idle ***********************************/
/* called by the message loop, and by paint messages                        */
void Idle( void )
{
    ConvertPropertyToStruct(FindSubPropertyByName(LocalTemplates, "global_properties"),
                            GlobalProperties,
                            &Globals);

    gravity.m_g.set_y(Globals.Gravity);
    
    PS2Controller.sample();
    Vec2 LeftStick = PS2Controller.left_stick();
    Vec2 RightStick = PS2Controller.right_stick();

    /**************************** step the simulator ****************************/   

    Game_Tick();

    if (spring)                             //update
        spring->m_a = Handle(world_body, world_mouse);

    if (!game_paused)
    {
       int const NumberOfSimSteps = 1;
       for(int StepCounter = 0;StepCounter < NumberOfSimSteps;++StepCounter) {
          // step as many times as you like
          simulator.step(1/(NumberOfSimSteps*GetFrameRate()));
          // do any game stuff you want here (like per-step collision response, etc.)
       }
    }
    // now wait for the rest of the time so we're consistent on fast machines
    WaitFrameRate(GetFrameRate());

    /**************************** render everything *****************************/
   
    // lots of gl stuff
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glDisable(GL_DEPTH_TEST);

    glLoadIdentity();
    glRotated(180.0*(-CameraAngleRadians/M_PI),0,0,1);
    glTranslatef(-CameraX,-CameraY,-CameraZ);

    if (draw_debugging_info)
    {
       // draw axes
       glBegin(GL_LINES);
       glColor3f(1,0,0);  glVertex2f(0,0);  glVertex2f(1,0);
       glColor3f(0,1,0);  glVertex2f(0,0);  glVertex2f(0,1);
       glEnd();
    }

    if (cur_game_state==kGameStateFreeForm)
    {
       // draw mouse
       glPointSize(4.0);
       glBegin(GL_POINTS);
       glColor3f(0,0,0);  glVertex2fv(world_mouse.m_v);
       glEnd();
    }

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    // iterate all the bodies, rendering them
    glColor3f(0,1,0);
    for(int i = 0;i < objs.size();++i) 
    {
        assert(objs[i]);
        game_object &obj = *objs[i];

        glBody1v(obj, obj.rb->m_Pbody.get(), true);
    }

    glColor3f(0,0,0);
    for(int i = 0;i < objs.size();++i) 
    {
        assert(objs[i]);
        game_object &obj = *objs[i];

        glBody1v(obj, obj.rb->m_Pbody.get(), false);
    }

    Game_Draw();
    
    // iterate all the forces, rendering them
    glColor3f(1,0,0);
    for(int i = 0;i < forces.size();++i)
    {
        assert(forces[i]);
        Binary_Force *Force = forces[i];

        glForce1v(Force);
    }
    glForce1v(spring); 

    glDisable(GL_BLEND);

    if (draw_debugging_info)
    {
       glColor3f(0, 0, 1);
       glBegin(GL_LINES);
       float StickScale = 10.0f;
       glVertex2f(0, 0);
       glVertex2f(StickScale * LeftStick.x(), StickScale * LeftStick.y());
       glVertex2f(0, 0);
       glVertex2f(StickScale * RightStick.x(), StickScale * RightStick.y());
       glEnd();
       
       glColor3f(0, 0, 0);
       char buf[100];
       sprintf(buf,"wm: %5.1f, %5.1f    sm: %5.1f, %5.1f",
               world_mouse.x(),world_mouse.y(),window_mouse.x(),window_mouse.y());
       DrawString(50,50,buf);
       
       char *ButtonStrings[] =
          {
             "BUTTON_RIGHT_D",
             "BUTTON_RIGHT_U",
             "BUTTON_RIGHT_L",
             "BUTTON_RIGHT_R",
             "BUTTON_SHOULDER_R1",
             "BUTTON_SHOULDER_R2",
             "BUTTON_SHOULDER_L1",
             "BUTTON_SHOULDER_L2",
             "BUTTON_RSTICK",
             "BUTTON_LSTICK",
             "BUTTON_START",
             "BUTTON_SELECT",
             "BUTTON_LEFT_D",
             "BUTTON_LEFT_U",
             "BUTTON_LEFT_L",
             "BUTTON_LEFT_R",
          };
       
       float Y = 90;
       for(int ButtonIndex = 0;
           ButtonIndex < Joystick::NUM_BUTTONS;
           ++ButtonIndex)
       {
          if(PS2Controller.button_is_down(ButtonIndex))
          {
             DrawString(50.f,Y,ButtonStrings[ButtonIndex]);
             Y += 10;
          }
       }
       
       DrawTimingBars();
    }
    SwapBuffers(WindowDC);
}

/******************************** AppWndProc ********************************/
/* lame to expose this, but the alternatives all suck too                   */
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
    switch (msg)
    {
        // keyboard
        case WM_CHAR:
            break;

        case WM_MOUSEWHEEL:
        {
            float d = float((short)HIWORD(wParam))/float(WHEEL_DELTA);
            CameraZ *= 1.0f + d/10.0f;
            world_mouse = WorldFromWindowCoords(window_mouse);
            break;
        }
        case WM_LBUTTONDOWN:
        {
            SetCapture(hwnd);

            Vec2 body_space_position;
            Rigid_Body* rb = simulator.pick(body_space_position, world_mouse);
            if (rb)
            {
                spring = new Spring_Force(
                    Handle(world_body, world_mouse),
                    Handle(*rb, body_space_position), 
                    Globals.MouseSpringKs,
                    Globals.MouseSpringKd,
                    Globals.MouseSpringRestLength
                    );
                simulator.add(*spring);
            }
        } break;

        case WM_LBUTTONUP:
        {
            if(!(wParam & MK_MBUTTON & MK_RBUTTON))
            {
                ReleaseCapture();
            }

            ClearMouseSpring();
        } break;

        case WM_RBUTTONDOWN:
        {
            SetCapture(hwnd);

            if(spring)
            {
                forces.push_back(spring);
                spring = 0;
            }
            break;
        }
        case WM_MBUTTONDOWN:
        {
            MovingCamera = true;
        } break;
        case WM_MBUTTONUP:
        {
            MovingCamera = false;
        } break;

        case WM_RBUTTONUP:
        {
            if(!(wParam & MK_MBUTTON & MK_LBUTTON))
            {
                ReleaseCapture();
            }
        } break;
            
        case WM_KEYDOWN: {
            switch(wParam) {
                case VK_LEFT: {
                    CameraAngleRadians += float((5.0/180.0)*M_PI);
                    break;
                }
                case VK_RIGHT: {
                    CameraAngleRadians -= float((5.0/180.0)*M_PI);
                    break;
                }

                case 'M': {
                    // demo the audio stuff
                    if(0x8000 & GetKeyState(VK_SHIFT)) {
                        music.play("background");
                    } else {
                        zsoundPlay("content/sounds/bad.0002.wav");
                    }
                    break;
                }

                case 'E': {
                    launchEditor(CurLevelFileName);
                    Game_Init();
                } break;

                case VK_ESCAPE: {
                    AppQuit = true;
                    break;
                }

                default: {
                   Game_Key(wParam);
                   break;
                }
            }
            break;
        }
        case WM_SYSKEYDOWN:
        case WM_SYSKEYUP:
        case WM_KEYUP:
            break;

        case WM_MOUSEMOVE: {
            Vec2 new_window_mouse = Vec2(float(GET_X_LPARAM(lParam)),float(GET_Y_LPARAM(lParam)));
            if(MovingCamera)
            {
                Vec2 d_world = WorldFromWindowCoords(new_window_mouse) - world_mouse;
                CameraX -= d_world.x();
                CameraY -= d_world.y();
            }
            window_mouse = new_window_mouse;
            world_mouse = WorldFromWindowCoords(window_mouse);
            break;
        }

         
            /**************** below is stuff you probably don't care about ****************/
         
        case WM_DESTROY: {
            PostQuitMessage(0);
            break;
        }
        case WM_PAINT: {
            PAINTSTRUCT ps;
            BeginPaint(hwnd,&ps);
            Idle();
            EndPaint(hwnd,&ps);
            return 0L;
        }
        case WM_SIZE: {
            Resize(LOWORD(lParam),HIWORD(lParam));
            break;
        }
    }
    return (LONG)DefWindowProc(hwnd,msg,(LONG)wParam,(LONG)lParam);
}

/********************************** Resize **********************************/
/* called on WM_SIZE, and at the bottom of initialize                       */
void Resize( int w, int h )
{
    // compute our frustrum planes, the x,y plane values are on the near clip plane
    WindowWidth = w;
    WindowHeight = h;
    AspectRatio = float(w)/float(h);
    float ymax = ZNear * float(tan(FOVY/2.0));
    float ymin = -ymax;
    float xmin = AspectRatio * ymin;
    float xmax = AspectRatio * ymax;
   
    // load the viewport and projection matrix, gl stuff
    glViewport(0,0,w,h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // gluPerspective doesn't set the near plane as the projection
    // plane, so you can't do good math, so do it ourselves
    glFrustum(xmin,xmax,ymin,ymax,ZNear,ZFar);
    glMatrixMode(GL_MODELVIEW);
}

/******************************* Uninitialize *******************************/
/* called before the graphics is destroyed, but the window is already toast */
void Uninitialize( void )
{
}

/********************************* WinMain **********************************/

int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int sw)
{
    // register and create our window
    WNDCLASS cls;
    if(!hPrev) {
        cls.hCursor        = LoadCursor(NULL,IDC_ARROW);
        cls.hIcon          = 0;
        cls.lpszMenuName   = 0;
        cls.lpszClassName  = AppName;
        cls.hbrBackground  = 0;
        cls.hInstance      = hInst;
        cls.style          = CS_VREDRAW | CS_HREDRAW | CS_DBLCLKS | CS_OWNDC;
        cls.lpfnWndProc    = (WNDPROC)AppWndProc;
        cls.cbWndExtra     = 0;
        cls.cbClsExtra     = 0;

        if (!RegisterClass(&cls))
            return FALSE;
    }

    DWORD Style = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
    RECT WindowRect = { 0, 0 };
    WindowRect.right = WindowWidth;
    WindowRect.bottom = WindowHeight;
    BOOL Menu = FALSE;

    AdjustWindowRect(&WindowRect,Style,Menu);
    
    WindowHandle = CreateWindow(AppName,AppName,Style,
                                0,50,
                                WindowRect.right-WindowRect.left,
                                WindowRect.bottom-WindowRect.top,
                                0,0,hInst,0);

    // create our gl stuff
    WindowDC = GetDC(WindowHandle);
    PIXELFORMATDESCRIPTOR FormatDescriptor = {
        sizeof (PIXELFORMATDESCRIPTOR), 1,
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA, 32,   // 32bpp
        0,0, 0,0, 0,0, 0,0,
        0, 0,0,0,0,
        24, 8,               // depth, stencil
        0,
        PFD_MAIN_PLANE,
        0,
        0,0,0
    };
    int FormatIndex = ChoosePixelFormat(WindowDC,&FormatDescriptor);
    SetPixelFormat(WindowDC,FormatIndex,&FormatDescriptor);
    WindowRC = wglCreateContext(WindowDC);
    wglMakeCurrent(WindowDC,WindowRC);

    // okay, ready to go
   
    Initialize();    // call the app
   
    ShowWindow(WindowHandle,sw);
   
    // pump the message queue
    MSG msg;
    while(!AppQuit) {
        if(PeekMessage(&msg,0,0,0,PM_REMOVE)) {
            if(msg.message == WM_QUIT)
                break;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        } else {
            Idle();    // call the app
        }
    }

    Uninitialize();  // call the app
   
    // clean up gl stuff and exit
    wglMakeCurrent(0,0);
    wglDeleteContext(WindowRC);
    ReleaseDC(WindowHandle,WindowDC);
   
    return int(msg.wParam);
}

/********************** utilities that use our globals **********************/
Vec2 WorldFromWindowCoords( Vec2 window )
{
    float xn = float(window.x())/WindowWidth - 0.5f;
    float yn = float(WindowHeight-window.y())/WindowHeight - 0.5f;
    float xc = float(xn*2.0*CameraZ*tan(FOVY/2)*AspectRatio);
    float yc = float(yn*2.0*CameraZ*tan(FOVY/2));
    return Vec2(CameraX,CameraY) + Matrix22::rotation(CameraAngleRadians) * Vec2(xc,yc);
}
Vec2 WindowFromWorldCoords( Vec2 world )
{
    Vec2 c = Matrix22::rotation(-CameraAngleRadians) * (world - Vec2(CameraX,CameraY));
    float xn = float(c.x()/(2.0*CameraZ*tan(FOVY/2)*AspectRatio));
    float yn = float(c.y()/(2.0*CameraZ*tan(FOVY/2)));
    float xs = WindowWidth*(xn + 0.5f);
    float ys = WindowHeight - WindowHeight*(yn + 0.5f);
    return Vec2(xs,ys);
}
